# -*- coding: utf-8 -*-
THIRD_PARTY_PACKAGES = (
    'Zope2',
    'ZODB3',
    'txtfilter',
    'Products.CMFActionIcons',
    'Products.CMFCalendar',
    'Products.CMFCore',
    'Products.CMFDefault',
    'Products.CMFTopic',
    'Products.CMFUid',
    'Products.DCWorkflow',
    'Products.GenericSetup',
    'Products.GroupUserFolder',
    'Products.PluggableAuthService',
    'Products.PluginRegistry',
    'Products.ZCatalog',
)

IGNORED_PACKAGES = (
    'plone.releaser',
)

ALWAYS_CHECKED_OUT = (
    'Plone',
    'Products.CMFPlone',
    'plone.app.upgrade',
    'plone.app.locales',
)

# Upon checking a package...
# ... ask every time if an action should be performed
ACTION_INTERACTIVE = 'interactive'
# ... don't ask anything and perform all actions
ACTION_BATCH = 'batch'
# ... don't ask anything *AND* don't make any action, just show the status
# of the package
ACTION_REPORT = 'report'

PACKAGE_ACTIONS = (
    ACTION_BATCH,
    ACTION_INTERACTIVE,
    ACTION_REPORT,
)



# -*- coding: utf-8 -*-
from argh import ArghParser
from argh import arg
from argh.decorators import named
from distutils.version import StrictVersion
from github import Github
from launchpadlib.launchpad import Launchpad
from plone.releaser import ACTION_BATCH
from plone.releaser import ACTION_INTERACTIVE
from plone.releaser import ACTION_REPORT
from plone.releaser import THIRD_PARTY_PACKAGES
from plone.releaser import pypi
from plone.releaser.buildout import Buildout
from plone.releaser.buildout import CheckoutsFile
from plone.releaser.buildout import VersionsFile
from plone.releaser.package import Package
from progress.bar import Bar

import datetime
import keyring


# TODO
buildout = Buildout()


def checkPypi(user):
    for package in buildout.sources:
        if package in THIRD_PARTY_PACKAGES:
            pass
        else:
            if not pypi.can_user_release_package_to_pypi(user, package):
                print "{0}: {1}".format(
                    package,
                    ', '.join(pypi.get_users_with_release_rights(package))
                )


@named('jenkins')
def jenkins_report():
    """Read-only version of checkAllPackagesForUpdates."""
    sources = buildout.sources
    for package_name, source in iter(sources.iteritems()):
        pkg = Package(buildout, package_name)
        pkg(action=ACTION_REPORT)


@arg('--interactive', default=False)
def checkPackageForUpdates(package_name, **kwargs):
    pkg = Package(buildout, package_name)
    if kwargs['interactive']:
        pkg(action=ACTION_INTERACTIVE)
    else:
        pkg(action=ACTION_BATCH)


@named('report')
@arg('--interactive', default=False)
def checkAllPackagesForUpdates(**kwargs):
    sources = buildout.sources
    for package_name, source in Bar('Scanning').iter(sources.iteritems()):
        pkg = Package(buildout, package_name)
        if kwargs['interactive']:
            pkg(action=ACTION_INTERACTIVE)
        else:
            pkg(action=ACTION_BATCH)


def pulls():
    client_id = 'b9f6639835b8c9cf462a'
    client_secret = keyring.get_password('plone.releaser', client_id)

    g = Github(client_id=client_id, client_secret=client_secret)

    for package_name, source in buildout.sources.iteritems():
        if source.path:
            repo = g.get_repo(source.path)
            pulls = [a for a in repo.get_pulls(
                'open') if a.head.ref == source.branch]
            if pulls:
                print package_name
                for pull in pulls:
                    print "    {0}: {1} ({2})".format(
                        pull.user.login,
                        pull.title,
                        pull.url
                    )


@named('changelog')
@arg('--start')
@arg('--end', default='here')
def changelog(**kwargs):
    from plone.releaser.changelog import build_unified_changelog
    build_unified_changelog(kwargs['start'], kwargs['end'])


@named('launchpad')
def create_launchpad_release(version):
    launchpad = Launchpad.login_with('plone.releaser', 'production')
    plone = launchpad.projects['plone']
    parsed_version = StrictVersion(version)
    # Blech. This feels flimsy
    series_name = '.'.join([str(a) for a in parsed_version.version[0:2]])
    series = plone.getSeries(name=series_name)
    if series is None:
        return "No series named {0}.".format(series_name)
    now = datetime.datetime.now().isoformat()
    milestone = series.newMilestone(name=version,
                                    date_targeted=now)
    # TODO: Get release notes
    release = milestone.createProductRelease(date_released=now,
                                             release_notes='')

    release_url = release.web_link

    return release_url


def check_checkout(package_name, path):
    if package_name not in CheckoutsFile(path):
        msg = 'Your package {0} is not on auto-checkout section'
        raise KeyError(msg.format(package_name))


def append_jenkins_build_number_to_package_version(jenkins_build_number):
    from zest.releaser.vcs import BaseVersionControl
    from zest.releaser.utils import cleanup_version
    vcs = BaseVersionControl()
    old_version = cleanup_version(vcs.version)
    new_version = '{0}.{1}'.format(old_version, jenkins_build_number)
    vcs.version = new_version
    return new_version


def set_package_version(version_file_path, package_name, new_version):
    versions = VersionsFile(version_file_path)
    versions.set(package_name, new_version)


class Manage(object):

    def __call__(self, **kwargs):
        parser = ArghParser()
        parser.add_commands(
            [checkPypi,
             checkPackageForUpdates,
             checkAllPackagesForUpdates,
             pulls,
             changelog,
             create_launchpad_release,
             check_checkout,
             append_jenkins_build_number_to_package_version,
             set_package_version,
             jenkins_report])
        parser.dispatch()


manage = Manage()



# -*- coding: utf-8 -*-
from argh.interaction import confirm
from configparser import NoOptionError
from contextlib import contextmanager
from plone.releaser import ACTION_BATCH
from plone.releaser import ACTION_INTERACTIVE
from plone.releaser import ACTION_REPORT
from plone.releaser import ALWAYS_CHECKED_OUT
from plone.releaser import IGNORED_PACKAGES
from plone.releaser import PACKAGE_ACTIONS
from plone.releaser import THIRD_PARTY_PACKAGES
from plone.releaser.db import IgnoresDB
from shutil import rmtree
from tempfile import mkdtemp

import git
import os


@contextmanager
def git_repo(source):
    """Handle temporal git repositories.

    It ensures that a git repository is cloned on a temporal folder that is
    removed after being used.

    See an example of this kind of context managers here:
    http://preshing.com/20110920/the-python-with-statement-by-example/
    """
    tmp_dir = mkdtemp()
    repo = git.Repo.clone_from(
        source.url,
        tmp_dir,
        branch=source.branch,
        depth=100
    )

    # give the control back
    yield repo

    # cleanup
    del repo
    rmtree(tmp_dir)


@contextmanager
def buildout_coredev():
    """Context manager for buildout.coredev git repositories.

    It ensures that the git repository is cloned on a temporal folder and that
    everything is cleaned up after being used.
    """
    repo = git.Repo(os.getcwd())
    yield repo
    del repo


class Package(object):

    # A reference to an plone.releaser.buildout.Buildout instance
    buildout = None

    # The package that's checked
    name = None

    # If the user that runs this script is expected to make interactive choices
    interactive = False

    # If only a report has to be produced and no changes should be pushed
    # anywhere, i.e. jenkins.plone.org creating reports for example
    report_only = False

    # The package URI as defined on sources.cfg
    source = None

    # The package version as found on versions.cfg
    version = None

    # Database of per package ignored commits
    commit_ignores = None

    def __init__(self, buildout, package):
        self.buildout = buildout
        self.name = package
        self.source = self.buildout.sources.get(self.name)
        self.version = self.get_version()
        self.commit_ignores = IgnoresDB()

    def __call__(self, action=ACTION_INTERACTIVE):
        if action not in PACKAGE_ACTIONS:
            print 'This package action does not exist: {0}'.format(action)
            return
        self.set_interaction_and_report(action)

        # exit early if the package:
        # - is on the ignored list
        # - there is no version available
        # - is not hosted on a git VCS
        if self.name in IGNORED_PACKAGES or \
                self.version is None or \
                not self.is_git_hosted():
            return

        # clone the package and gather data about it
        with git_repo(self.source) as repo:
            # exit early if no tag can be found
            latest_tag_in_branch = self.latest_tag(repo)
            if latest_tag_in_branch is None:
                return

            # if there is a newer tag of the package not in buildout.coredev
            # versions.cfg, ask/add/report about it
            self.update_version(latest_tag_in_branch)

            commits_since_release = self.latest_commits(repo)
            latest_ignored_commit = self.commit_ignores.get(self.name)
            commits_since_ignore = None
            if latest_ignored_commit is not None:
                commits_since_ignore = self._commits_between(
                    repo,
                    latest_ignored_commit,
                    self.source.branch
                )

            # if there are no changes since the last release (i.e. last tag)
            if not commits_since_release \
                    or 'Back to development' in commits_since_release[0].message \
                    or commits_since_release[0].message.startswith('vb'):

                self.remove()

            elif commits_since_ignore is None:
                # Check for checkout
                if self.name not in self.buildout.checkouts:
                    msg = '\nWARNING: No auto-checkout exists for {0}\n Changes in {0}:'  # noqa
                    self.print_commits(
                        commits_since_release,
                        message=msg.format(self.name)
                    )

                    if self.name in THIRD_PARTY_PACKAGES:
                        msg = 'NOTE: {0} is a third-party package.'
                        print msg.format(self.name)

                    self.add(commits_since_release)

                else:
                    if not self.interactive:
                        msg = '\nChanges in {0}:'.format(self.name)
                        self.print_commits(commits_since_release, message=msg)

                        if self.name in THIRD_PARTY_PACKAGES:
                            msg = 'NOTE: {0} is a third-party package.'
                            print msg.format(self.name)

    def set_interaction_and_report(self, action):
        if action == ACTION_REPORT:
            self.interactive = False
            self.report_only = True
        elif action == ACTION_INTERACTIVE:
            self.interactive = True
            self.report_only = False
        elif action == ACTION_BATCH:
            self.interactive = False
            self.report_only = False

    def is_git_hosted(self):
        if self.source.protocol != 'git':
            if self.report_only:
                msg = 'Skipped check of {0} as it\'s not a git repo.'
                print msg.format(self.name)
            return False
        return True

    def get_version(self):
        version = None
        try:
            version = self.buildout.get_version(self.name)
        except (NoOptionError, KeyError):
            if self.report_only:
                print 'No version available for {0}'.format(self.name)

        return version

    def latest_tag(self, repo):
        tag = None
        try:
            tag = repo.git.describe('--abbrev=0', '--tags')
        except git.exc.GitCommandError:
            if self.report_only:
                print 'Unable to check tags for {0}'.format(self.name)

        return tag

    def latest_commits(self, repo):
        commits = None

        try:
            commits = self._commits_between(
                repo,
                self.version,
                self.source.branch
            )
        except git.exc.GitCommandError:
            print '\nCould not read commits for package {0}'.format(self.name)

        return commits

    @staticmethod
    def _commits_between(repo, start, end):
        return list(
            repo.iter_commits(
                '{0}..{1}'.format(start, end)
            )
        )

    def remove(self):
        if self.name in self.buildout.checkouts and \
                self.name not in ALWAYS_CHECKED_OUT:
            msg = '\nNo new changes in {0}, but it is listed for auto-checkout.'  # noqa
            print msg.format(self.name)

        if self.report_only:
            return

        msg = 'Remove {0} from checkouts.cfg'.format(self.name)
        if confirm(msg, default=True, skip=not self.interactive):
            self.buildout.remove_from_checkouts(self.name)

            with buildout_coredev() as core_repo:
                checkouts_path = os.path.join(os.getcwd(), 'checkouts.cfg')
                core_repo.git.add(checkouts_path)
                msg = 'No new changes in {0}'.format(self.name)
                core_repo.git.commit(message=msg)

    def add(self, commits_since_release):
        if self.report_only:
            return

        msg = 'Add {0} to checkouts.cfg'.format(self.name)
        if confirm(msg, default=True, skip=not self.interactive):
            self.buildout.add_to_checkouts(self.name)

            with buildout_coredev() as core_repo:
                checkouts_path = os.path.join(os.getcwd(), 'checkouts.cfg')
                core_repo.git.add(checkouts_path)
                msg = '{0} has changes.'.format(self.name)
                core_repo.git.commit(message=msg)

        elif confirm('Ignore changes in  {0}'.format(self.name),
                     default=False,
                     skip=not self.interactive):
            self.commit_ignores.set(
                self.name,
                commits_since_release[0].hexsha
            )

    @staticmethod
    def print_commits(commits_list, message=None):
        if message:
            print message

        for commit in commits_list:
            print '    {0}: {1}'.format(
                commit.author.name.encode('ascii', 'replace'),
                commit.summary.encode('ascii', 'replace')
            )

    def update_version(self, tag):
        if tag <= self.version:
            return

        msg = '\nNewer version {0} is available for {1} (Currently {2})'
        print msg.format(tag, self.name, self.version)

        if self.report_only:
            return

        if confirm('Update versions.cfg',
                   default=True,
                   skip=not self.interactive):
            self.buildout.set_version(self.name, tag)

            with buildout_coredev() as core_repo:
                versions_path = os.path.join(os.getcwd(), 'versions.cfg')
                core_repo.git.add(versions_path)
                core_repo.git.commit(message='{0}={1}'.format(self.name, tag))
                core_repo.git.push()



# -*- coding: utf-8 -*-
from plone.releaser.buildout import CheckoutsFile
from plone.releaser.buildout import VersionsFile
from plone.releaser.changelog import Changelog
from plone.releaser.pypi import can_user_release_package_to_pypi
from zest.releaser import pypi
from zest.releaser.utils import ask

import git
import os
import sys


def check_pypi_access(data):
    pypi_user = pypi.PypiConfig().config.get('pypi', 'username')
    if not can_user_release_package_to_pypi(pypi_user, data['name']):
        msg = "User {0} does not have pypi release rights to {1}. Continue?"
        if not ask(msg.format(pypi_user, data['name']), default=False):
            sys.exit()


def show_changelog_entries(data):
    # Find changelog
    # TODO: Figure out how to catch malformed rst
    if data['history_file'] is not None:
        changelog = Changelog(file_location=data['history_file'])
    # Get top release's entry
    entries = changelog.latest()
    if entries is None:
        if not ask("Unable to parse changelog. Continue?", default=True):
            sys.exit()
        return
    print "Changelog entries for version {0}.".format(data['new_version'])
    for entry in entries:
        if isinstance(entry, list):
            print '\n'.join(entry)
        else:
            print entry
    if not ask("Continue?", default=True):
        sys.exit()


def update_core(data):
    if ask("Ok to update coredev versions.cfg/checkouts.cfg?", default=True):
        root_path = os.path.join(os.getcwd(), '../../')
        g = git.Git(root_path)
        g.pull()  # make sure buildout.coredev is up-to-date
        package_name = data['name']
        new_version = data['version']
        update_versions(package_name, new_version)
        update_checkouts(package_name)
        # git commit
        message = "{0} {1}".format(package_name, new_version)
        g.add('versions.cfg')
        g.add('checkouts.cfg')
        print "Committing changes."
        g.commit(message=message)
        print "Pushing changes to server."
        g.push()


def update_versions(package_name, new_version):
    # Update version
    print "Updating versions.cfg"
    path = os.path.join(os.getcwd(), '../../versions.cfg')
    versions = VersionsFile(path)
    versions.set(package_name, new_version)


def update_checkouts(package_name):
    # Remove from checkouts.cfg
    print "Removing package from checkouts.cfg"
    path = os.path.join(os.getcwd(), '../../checkouts.cfg')
    checkouts = CheckoutsFile(path)
    checkouts.remove(package_name)



# -*- coding: utf-8 -*-
from UserDict import UserDict
from collections import OrderedDict
from configparser import ConfigParser
from configparser import ExtendedInterpolation

import os
import re


PATH_RE = re.compile(
    '(\w+://)(.+@)*([\w\d\.]+)(:[\d]+){0,1}/(?P<path>.+(?=\.git))(\.git)'
)


class Source(object):

    def __init__(self, protocol=None, url=None, push_url=None, branch=None):
        self.protocol = protocol
        self.url = url
        self.push_url = push_url
        self.branch = branch

    def create_from_string(self, source_string):
        protocol, url, extra_1, extra_2, extra_3 = (
            lambda a, b, c=None, d=None, e=None: (a, b, c, d, e))(*source_string.split())
        for param in [extra_1, extra_2, extra_3]:
            if param is not None:
                key, value = param.split('=')
                setattr(self, key, value)
        self.protocol = protocol
        self.url = url
        if self.push_url is not None:
            self.push_url = self.push_url.split('=')[-1]
        if self.branch is None:
            self.branch = 'master'
        else:
            self.branch = self.branch.split('=')[-1]
        return self

    @property
    def path(self):
        if self.url:
            match = PATH_RE.match(self.url)
            if match:
                return match.groupdict()['path']
        return None


class VersionsFile(object):

    def __init__(self, file_location):
        self.file_location = file_location

    @property
    def versions(self):
        config = ConfigParser(interpolation=ExtendedInterpolation())
        with open(self.file_location) as f:
            config.read_file(f)
        return config['versions']

    def __contains__(self, package_name):
        return package_name.lower() in self.versions.keys()

    def __getitem__(self, package_name):
        if self.__contains__(package_name):
            return self.versions.get(package_name)
        else:
            raise KeyError

    def __setitem__(self, package_name, new_version):
        path = os.path.join(os.getcwd(), self.file_location)
        with open(path, 'r') as f:
            versionstxt = f.read()

        if package_name not in self:
            newline = "{0} = {1}".format(package_name, new_version)
            versionstxt += newline

        reg = re.compile(
            "(^{0}[\s\=]+)[0-9\.abrc]+".format(package_name),
            re.MULTILINE
        )
        newVersionsTxt = reg.sub(r"\g<1>{0}".format(new_version), versionstxt)
        with open(path, 'w') as f:
            f.write(newVersionsTxt)

    def get(self, package_name):
        return self.__getitem__(package_name)

    def set(self, package_name, new_version):
        return self.__setitem__(package_name, new_version)


class SourcesFile(UserDict):

    def __init__(self, file_location):
        self.file_location = file_location

    @property
    def data(self):
        config = ConfigParser(interpolation=ExtendedInterpolation())
        config.optionxform = str
        with open(self.file_location) as f:
            config.read_file(f)
        sources_dict = OrderedDict()
        for name, value in config['sources'].items():
            source = Source().create_from_string(value)
            sources_dict[name] = source
        return sources_dict

    def __setitem__(self, package_name, value):
        raise NotImplementedError

    def __iter__(self):
        return self.data.__iter__()


class CheckoutsFile(UserDict):

    def __init__(self, file_location):
        self.file_location = file_location

    @property
    def data(self):
        config = ConfigParser(interpolation=ExtendedInterpolation())
        with open(self.file_location) as f:
            config.read_file(f)
        checkouts = config.get('buildout', 'auto-checkout')
        checkout_list = checkouts.split('\n')
        return checkout_list

    def __contains__(self, package_name):
        return package_name in self.data

    def __setitem__(self, package_name, enabled=True):
        path = os.path.join(os.getcwd(), self.file_location)
        with open(path, 'r') as f:
            checkoutstxt = f.read()
        with open(path, 'w') as f:
            if enabled:
                fixes_text = "# Test fixes only"
                reg = re.compile(
                    "^[\s]*{0}\n".format(fixes_text),
                    re.MULTILINE
                )
                newCheckoutsTxt = reg.sub(
                    '    {0}\n{1}\n'.format(package_name, fixes_text),
                    checkoutstxt
                )
            else:
                reg = re.compile(
                    "^[\s]*{0}\n".format(package_name),
                    re.MULTILINE
                )
                newCheckoutsTxt = reg.sub('', checkoutstxt)
            f.write(newCheckoutsTxt)

    def __delitem__(self, package_name):
        return self.__setitem__(package_name, False)

    # def setAutoCheckouts(self, checkouts_list):
    #     config = ConfigParser(interpolation=ExtendedInterpolation())
    #     with open(self.file_location) as f:
    #         config.read_file(f)
    #     checkouts = '\n'.join(checkouts_list)
    #     config.set('buildout', 'auto-checkout', checkouts)
    #     with open(self.file_location, 'w') as f:
    #         config.write(f)

    def add(self, package_name):
        # TODO: Handle test-fix-only as well
        return self.__setitem__(package_name, True)

    def remove(self, package_name):
        # Remove from checkouts.cfg
        return self.__delitem__(package_name)


class Buildout(object):

    def __init__(self,
                 sources_file='sources.cfg',
                 checkouts_file='checkouts.cfg',
                 versions_file='versions.cfg'):
        self.sources = SourcesFile(sources_file)
        self.versions = VersionsFile(versions_file)
        self.checkouts = CheckoutsFile(checkouts_file)

    def add_to_checkouts(self, package_name):
        return self.checkouts.add(package_name)

    def remove_from_checkouts(self, package_name):
        return self.checkouts.remove(package_name)

    def get_version(self, package_name):
        return self.versions.get(package_name)

    def set_version(self, package_name, new_version):
        return self.versions.set(package_name, new_version)



# -*- coding: utf-8 -*-
import json
import os


class IgnoresDB(object):

    def __init__(self):
        self._filename = '.package_ignores'
        if not os.path.isfile(self._filename):
            open(self._filename, 'w').close()

        with open(self._filename, 'r') as f:
            content = f.read()
            if content != '':
                self._db = json.loads(content)
            else:
                self._db = {}

    def save(self):
        with open(self._filename, 'w+') as f:
            f.write(json.dumps(self._db))

    def get(self, package_name):
        return self._db.get(package_name)

    def set(self, package_name, sha):
        self._db[package_name] = sha
        self.save()

    def delete(self, package_name):
        del self._db[package_name]
        self.save()



# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup


version = '1.3'

long_description = '{0}\n{1}'.format(
    open('README.rst').read(),
    open('CHANGES.rst').read()
)

setup(
    name='plone.releaser',
    version=version,
    description='Plone release management utilities',
    long_description=long_description,
    # Get more strings from
    # http://pypi.python.org/pypi?:action=list_classifiers
    classifiers=[
      'Framework :: Plone',
      'Programming Language :: Python',
      'Programming Language :: Python :: 2.7',
    ],
    keywords='plone release',
    author='Eric Steele',
    author_email='eric@esteele.net',
    url='https://github.com/plone/plone.releaser',
    license='GPL',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['plone'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'argh',
        'gitpython>=0.3',
        'configparser',
        'argcomplete',
        'progress',
        'PyGithub',
        'keyring',
        'zest.releaser',
        'docutils',
        'launchpadlib',
    ],
    entry_points={
        'console_scripts': [
            'manage = plone.releaser.manage:manage',
        ],
        'zest.releaser.prereleaser.middle': [
            'show_changelog=plone.releaser.release:show_changelog_entries',
            'check_pypi=plone.releaser.release:check_pypi_access',
        ],
        'zest.releaser.releaser.after': [
            'update_core=plone.releaser.release:update_core',
        ]
    },
)



# -*- coding: utf-8 -*-
from collections import OrderedDict
from distutils.version import LooseVersion
from docutils.core import publish_doctree
from itertools import product
from plone.releaser.buildout import Buildout

import urllib


DIST_URL = "http://dist.plone.org/release/{0}/versions.cfg"

buildout = Buildout()


def pull_versions(version_number):
    package_versions = OrderedDict()
    if version_number == 'here':
        url = 'versions.cfg'
        versions_file = open(url)
    else:
        url = DIST_URL.format(version_number)
        versions_file = urllib.urlopen(url)
    if versions_file.code == 404:
        raise ValueError("Version %s not found." % version_number)
    for line in versions_file:
        line = line.strip().replace(" ", "")
        if line and not (line.startswith('#') or line.startswith('[')):
            try:
                package, version = line.split("=")
                version = LooseVersion(version)
            except ValueError:
                pass
            else:
                package_versions[package] = version
    print "Parsed {0}".format(url)
    return package_versions


def get_source_location(package_name):
    source = buildout.sources.get(package_name)
    if source is not None:
        # Go from this:
        # git://github.com/plone/plone.batching.git
        # to this:
        # https://raw.githubusercontent.com/plone/plone.batching
        url = source.url
        url = url.replace('git:', 'https:')
        url = url.replace('.git', '')
        url = url.replace('github.com', 'raw.githubusercontent.com')
        return url, source.branch
    return "", ""


def get_changelog(package_name):
    source_url, branch = get_source_location(package_name)
    if not source_url:
        return ''
    file_names = ['CHANGES', 'HISTORY']
    file_extensions = ['.rst', '.txt']
    if 'github' in source_url:
        paths = [
            '{0}/'.format(branch),
            '{0}/docs/'.format(branch),
        ]
    else:
        paths = ['/', '/docs/', '/'.join(package_name.split('.')) + '/']
    for pathable in product(paths, file_names, file_extensions):
        structure = ''.join(pathable)
        url = "{0}/{1}".format(source_url, structure)
        try:
            response = urllib.urlopen(url)
        except IOError:
            print "Unable to reach {0}".format(url)
        else:
            if response.code == 200:
                return response.read()
    return ''


class Changelog(object):

    def __init__(self, file_location=None, content=None):
        self.data = OrderedDict()
        if content is not None:
            self._parse(content)
        elif file_location is not None:
            with open(file_location, 'r') as f:
                self._parse(f.read())

    def __iter__(self):
        return self.data.__iter__()

    def iteritems(self):
        return self.data.iteritems()

    def get(self, version):
        return self.data.get(version)

    def get_changes(self, start_version, end_version=None):
        versions = self.data.keys()

        end_version_index = 0
        if end_version is not None:
            try:
                end_version_index = versions.index(str(end_version))
            except ValueError:
                raise ValueError("Unknown version {0}".format(end_version))
        try:
            start_version_index = versions.index(str(start_version))
        except ValueError:
            raise ValueError("Unknown version {0}".format(start_version))

        newer_releases = versions[end_version_index:start_version_index]
        changes = []
        for release in newer_releases:
            changes.extend(self.data[release])
        return changes

    def latest(self):
        if self.data.items():
            return self.data.items()[0]
        return None

    def _parse(self, content):
        tree = publish_doctree(content)

        def is_valid_version_section(x):
            if x.tagname == "section":
                try:
                    LooseVersion(x['names'][0].split()[0])
                except (ValueError, IndexError):
                    pass
                else:
                    return True
            return False

        def is_list_item(x):
            return x.tagname == 'list_item'

        found_sections = tree.traverse(condition=is_valid_version_section)
        for section in found_sections:
            version = section['names'][0].split()[0]
            list_items = section.traverse(condition=is_list_item)
            entries = [a.rawsource for a in list_items]
            self.data[version] = entries


def build_unified_changelog(start_version, end_version):
    try:
        prior_versions = pull_versions(start_version)
        current_versions = pull_versions(end_version)
    except ValueError, e:
        print e
        return

    output_str = ""
    try:
        for package, version in current_versions.iteritems():
            if package in prior_versions:
                prior_version = prior_versions[package]
                if version > prior_version:
                    print "{0} has a newer version".format(package)
                    packageChange = u"{0}: {1} {2} {3}".format(
                        package,
                        prior_version,
                        u"\u2192",
                        version
                    )
                    output_str += u"\n" + packageChange + \
                        u"\n" + u"-" * len(packageChange) + "\n"

                    logtext = get_changelog(package)
                    if not logtext:
                        print "No changelog found."
                        continue
                    changelog = Changelog(content=logtext)
                    try:
                        changes = changelog.get_changes(prior_version, version)
                    except ValueError, e:
                        print e
                    else:
                        for change in changes:
                            bullet = "- "
                            change = change.replace("\n", "\n" + " " * len(bullet))
                            output_str += bullet + change + u"\n"
    except KeyboardInterrupt:
        pass

    print output_str.encode('utf-8')



# -*- coding: utf-8 -*-
import xmlrpclib


def get_users_with_release_rights(package_name):
    client = xmlrpclib.ServerProxy('http://pypi.python.org/pypi')
    existing_admins = set([
        user for role, user in client.package_roles(package_name)])
    return existing_admins


def can_user_release_package_to_pypi(user, package_name):
    return user in get_users_with_release_rights(package_name)
